import numpy as np
from numba import jit, njit


@njit
def center_values(vals, mu, var):
    out = np.zeros_like(vals)

    for i in range(len(vals)):
        std = var[i]**0.5
        if std == 0:
            out[i] = 0
        else:
            out[i] = (vals[i] - mu[i])/std

    return out


@njit
def neighbor_smoothing(vals, neighbors, weights, _lambda=.9):
    """

    output is (neighborhood average) * _lambda + self * (1-_lambda)


    vals: expression matrix (genes x cells)
    neighbors: neighbor indices (cells x K)
    weights: neighbor weights (cells x K)
    _lambda: ratio controlling self vs. neighborhood
    """

    out = np.zeros_like(vals, dtype=np.float64)

    G = vals.shape[0]       # Genes

    for g in range(G):

        row_vals = vals[g, :]
        smooth_row_vals = neighbor_smoothing_row(
            row_vals, neighbors, weights, _lambda)

        out[g, :] = smooth_row_vals

    return out


@njit
def neighbor_smoothing_row(vals, neighbors, weights, _lambda=.9):
    """

    output is (neighborhood average) * _lambda + self * (1-_lambda)


    vals: expression matrix (genes x cells)
    neighbors: neighbor indices (cells x K)
    weights: neighbor weights (cells x K)
    _lambda: ratio controlling self vs. neighborhood
    """

    out = np.zeros_like(vals, dtype=np.float64)
    out_denom = np.zeros_like(vals, dtype=np.float64)

    N = neighbors.shape[0]  # Cells
    K = neighbors.shape[1]  # Neighbors

    for i in range(N):

        xi = vals[i]

        for k in range(K):
            j = neighbors[i, k]

            wij = weights[i, k]
            xj = vals[j]

            out[i] += xj*wij
            out[j] += xi*wij

            out_denom[i] += wij
            out_denom[j] += wij

    out /= out_denom

    out = (out * _lambda) + (1 - _lambda) * vals

    return out



from sklearn.neighbors import NearestNeighbors
import numpy as np
import pandas as pd
from math import ceil
from numba import jit
from tqdm import tqdm
from pynndescent import NNDescent


def neighbors_and_weights(data, n_neighbors=30, neighborhood_factor=3, approx_neighbors=True):
    """
    Computes nearest neighbors and associated weights for data
    Uses euclidean distance between rows of `data`

    Parameters
    ==========
    data: pandas.Dataframe num_cells x num_features

    Returns
    =======
    neighbors:      pandas.Dataframe num_cells x n_neighbors
    weights:  pandas.Dataframe num_cells x n_neighbors

    """

    coords = data.values

    if approx_neighbors:
        index = NNDescent(coords, n_neighbors=n_neighbors)
        ind, dist = index.neighbor_graph
    else:
        nbrs = NearestNeighbors(n_neighbors=n_neighbors,
                        algorithm="ball_tree").fit(coords)
        dist, ind = nbrs.kneighbors()

    weights = compute_weights(
        dist, neighborhood_factor=neighborhood_factor)

    ind = pd.DataFrame(ind, index=data.index)
    neighbors = ind
    weights = pd.DataFrame(weights, index=neighbors.index,
                           columns=neighbors.columns)

    return neighbors, weights


def neighbors_and_weights_from_distances(
    distances, n_neighbors=30, neighborhood_factor=3
):
    """
    Computes nearest neighbors and associated weights using
    provided distance matrix directly

    Parameters
    ==========
    distances: pandas.Dataframe num_cells x num_cells

    Returns
    =======
    neighbors:      pandas.Dataframe num_cells x n_neighbors
    weights:  pandas.Dataframe num_cells x n_neighbors

    """

    nbrs = NearestNeighbors(
        n_neighbors=n_neighbors, algorithm="brute", metric="precomputed"
    ).fit(distances.values)
    dist, ind = nbrs.kneighbors()

    weights = compute_weights(dist, neighborhood_factor=neighborhood_factor)

    ind = pd.DataFrame(ind, index=distances.index)
    neighbors = ind
    weights = pd.DataFrame(
        weights, index=neighbors.index, columns=neighbors.columns
    )

    return neighbors, weights


def compute_weights(distances, neighborhood_factor=3):
    """
    Computes weights on the nearest neighbors based on a
    gaussian kernel and their distances

    Kernel width is set to the num_neighbors / neighborhood_factor's distance

    distances:  cells x neighbors ndarray
    neighborhood_factor: float

    returns weights:  cells x neighbors ndarray

    """

    radius_ii = ceil(distances.shape[1] / neighborhood_factor)

    sigma = distances[:, [radius_ii-1]]
    sigma[sigma == 0] = 1

    weights = np.exp(-1 * distances**2 / sigma**2)

    wnorm = weights.sum(axis=1, keepdims=True)
    wnorm[wnorm == 0] = 1.0
    weights = weights / wnorm

    return weights


@jit(nopython=True)
def compute_node_degree(neighbors, weights):

    D = np.zeros(neighbors.shape[0])

    for i in range(neighbors.shape[0]):
        for k in range(neighbors.shape[1]):

            j = neighbors[i, k]
            w_ij = weights[i, k]

            D[i] += w_ij
            D[j] += w_ij

    return D


@jit(nopython=True)
def make_weights_non_redundant(neighbors, weights):
    w_no_redundant = weights.copy()

    for i in range(neighbors.shape[0]):
        for k in range(neighbors.shape[1]):
            j = neighbors[i, k]

            if j < i:
                continue

            for k2 in range(neighbors.shape[1]):
                if neighbors[j, k2] == i:
                    w_ji = w_no_redundant[j, k2]
                    w_no_redundant[j, k2] = 0
                    w_no_redundant[i, k] += w_ji

    return w_no_redundant

# Neighbors and weights given an ete3 tree instead

def _search(current_node, previous_node, distance):

    if current_node.is_root():
        nodes_to_search = current_node.children
    else:
        nodes_to_search = current_node.children + [current_node.up]
    nodes_to_search = [x for x in nodes_to_search if x != previous_node]

    if len(nodes_to_search) == 0:
        return {current_node.name: distance}

    result = {}
    for new_node in nodes_to_search:

        res = _search(new_node, current_node, distance+1)
        for k, v in res.items():
            result[k] = v

    return result


def _knn(leaf, K):

    dists = _search(leaf, None, 0)
    dists = pd.Series(dists)
    dists = dists + np.random.rand(len(dists)) * .9  # to break ties randomly

    neighbors = dists.sort_values().index[0:K].tolist()

    return neighbors


def tree_neighbors_and_weights(tree, n_neighbors, cell_labels):
    """
    Computes nearest neighbors and associated weights for data
    Uses distance along the tree object

    Names of the leaves of the tree must match the columns in counts

    Parameters
    ==========
    tree: ete3.TreeNode
        The root of the tree
    n_neighbors: int
        Number of neighbors to find
    cell_labels
        Labels of cells (barcodes)

    Returns
    =======
    neighbors:      pandas.Dataframe num_cells x n_neighbors
    weights:  pandas.Dataframe num_cells x n_neighbors

    """

    K = n_neighbors

    all_leaves = []
    for x in tree:
        if x.is_leaf():
            all_leaves.append(x)

    all_neighbors = {}

    for leaf in tqdm(all_leaves):
        neighbors = _knn(leaf, K)
        all_neighbors[leaf.name] = neighbors

    cell_ix = {c: i for i, c in enumerate(cell_labels)}

    knn_ix = np.zeros((len(all_neighbors), K), dtype='int64')
    for cell in all_neighbors:
        row = cell_ix[cell]
        nn_ix = [cell_ix[x] for x in all_neighbors[cell]]
        knn_ix[row, :] = nn_ix

    neighbors = pd.DataFrame(knn_ix, index=cell_labels)
    weights = pd.DataFrame(
        np.ones_like(neighbors, dtype='float64'),
        index=cell_labels
    )

    return neighbors, weights



import numpy as np
import pandas as pd
import warnings

from scipy.sparse import issparse, csr_matrix

from .knn import (
    neighbors_and_weights,
    neighbors_and_weights_from_distances,
    tree_neighbors_and_weights,
    make_weights_non_redundant,
)
from .local_stats import compute_hs
from .local_stats_pairs import (
    compute_hs_pairs, compute_hs_pairs_centered_cond)

from . import modules
from .plots import local_correlation_plot
from tqdm import tqdm


class Hotspot:

    def __init__(
            self, adata, layer_key=None, model='danb',
            latent_obsm_key=None, distances_obsp_key=None, tree=None,
            umi_counts_obs_key=None):
        """Initialize a Hotspot object for analysis

        Either `latent` or `tree` or `distances` is required.

        Parameters
        ----------
        adata : anndata.AnnData
            Count matrix (shape is cells by genes)
        layer_key: str
            Key in adata.layers with count data, uses adata.X if None.
        model : string, optional
            Specifies the null model to use for gene expression.
            Valid choices are:

                - 'danb': Depth-Adjusted Negative Binomial
                - 'bernoulli': Models probability of detection
                - 'normal': Depth-Adjusted Normal
                - 'none': Assumes data has been pre-standardized

        latent_obsm_key : string, optional
            Latent space encoding cell-cell similarities with euclidean
            distances.  Shape is (cells x dims). Input is key in adata.obsm
        distances_obsp_key : pandas.DataFrame, optional
            Distances encoding cell-cell similarities directly
            Shape is (cells x cells). Input is key in adata.obsp
        tree : ete3.coretype.tree.TreeNode
            Root tree node.  Can be created using ete3.Tree
        umi_counts_obs_key : str
            Total umi count per cell.  Used as a size factor.
            If omitted, the sum over genes in the counts matrix is used
        """
        counts = self._counts_from_anndata(adata, layer_key)
        distances = adata.obsp[distances_obsp_key] if distances_obsp_key is not None else None
        latent = adata.obsm[latent_obsm_key] if latent_obsm_key is not None else None
        umi_counts = adata.obs[umi_counts_obs_key] if umi_counts_obs_key is not None else None

        if latent is None and distances is None and tree is None:
            raise ValueError("Neither `latent_obsm_key` or `tree` or `distances_obsp_key` arguments were supplied.  One of these is required")

        if latent is not None and distances is not None:
            raise ValueError("Both `latent_obsm_key` and `distances_obsp_key` provided - only one of these should be provided.")

        if latent is not None and tree is not None:
            raise ValueError("Both `latent_obsm_key` and `tree` provided - only one of these should be provided.")

        if distances is not None and tree is not None:
            raise ValueError("Both `distances_obsp_key` and `tree` provided - only one of these should be provided.")

        if distances is not None:
            assert not issparse(distances)
            distances = pd.DataFrame(distances, index=adata.obs_names, columns=adata.obs_names)

        if latent is not None:
            latent = pd.DataFrame(latent, index=adata.obs_names)

        # because of transpose we check if its csr
        if issparse(counts) and not isinstance(counts, csr_matrix):
            warnings.warn("Hotspot will work faster when counts are a csc sparse matrix.")

        if tree is not None:
            try:
                all_leaves = []
                for x in tree:
                    if x.is_leaf():
                        all_leaves.append(x.name)
            except:
                raise ValueError("Can't parse supplied tree")

            if (
                len(all_leaves) != counts.shape[1] or
                len(set(all_leaves) & set(adata.obs_names)) != len(all_leaves)
               ):
                raise ValueError("Tree leaf labels don't match columns in supplied counts matrix")

        if umi_counts is None:
            umi_counts = counts.sum(axis=0)
            # handles sparse matrix outputs of sum
            umi_counts = np.asarray(umi_counts).ravel()
        else:
            assert umi_counts.size == counts.shape[1]

        if not isinstance(umi_counts, pd.Series):
            umi_counts = pd.Series(umi_counts, index=adata.obs_names)

        valid_models = {'danb', 'bernoulli', 'normal', 'none'}
        if model not in valid_models:
            raise ValueError(
                'Input `model` should be one of {}'.format(valid_models)
            )

        valid_genes = counts.sum(axis=1) > 0
        n_invalid = counts.shape[0] - valid_genes.sum()
        if n_invalid > 0:
            raise ValueError(
                "\nDetected all zero genes. Please filter adata and reinitialize."
            )

        self.adata = adata
        self.layer_key = layer_key

        self.counts = counts
        self.latent = latent
        self.distances = distances
        self.tree = tree
        self.model = model

        self.umi_counts = umi_counts

        self.graph = None
        self.modules = None
        self.local_correlation_z = None
        self.linkage = None
        self.module_scores = None

    @staticmethod
    def _counts_from_anndata(adata, layer_key, dense=False, pandas=False):
        counts = adata.layers[layer_key] if layer_key is not None else adata.X
        is_sparse = issparse(counts)
        # handles adata view
        # as sparse matrix in view is just a sparse matrix, while dense is ArrayView
        if not issparse(counts):
            counts = np.asarray(counts)
        counts = counts.transpose()

        if dense:
            counts = counts.A if is_sparse else counts
            is_sparse = False
        if pandas and is_sparse:
            raise ValueError("Set dense=True to return pandas output")
        if pandas and not is_sparse:
            counts = pd.DataFrame(counts, index=adata.var_names, columns=adata.obs_names)            

        return counts

    def create_knn_graph(
            self, weighted_graph=False, n_neighbors=30, neighborhood_factor=3, approx_neighbors=True):
        """Create's the KNN graph and graph weights

        The resulting matrices containing the neighbors and weights are
        stored in the object at `self.neighbors` and `self.weights`

        Parameters
        ----------
        weighted_graph: bool
            Whether or not to create a weighted graph
        n_neighbors: int
            Neighborhood size
        neighborhood_factor: float
            Used when creating a weighted graph.  Sets how quickly weights decay
            relative to the distances within the neighborhood.  The weight for
            a cell with a distance d will decay as exp(-d/D) where D is the distance
            to the `n_neighbors`/`neighborhood_factor`-th neighbor.
        approx_neighbors: bool
            Use approximate nearest neighbors or exact scikit-learn neighbors. Only
            when hotspot initialized with `latent`.
        """

        if self.latent is not None:
            neighbors, weights = neighbors_and_weights(
                self.latent, n_neighbors=n_neighbors,
                neighborhood_factor=neighborhood_factor,
                approx_neighbors=approx_neighbors)
        elif self.tree is not None:
            if weighted_graph:
                raise ValueError("When using `tree` as the metric space, `weighted_graph=True` is not supported")
            neighbors, weights = tree_neighbors_and_weights(
                self.tree, n_neighbors=n_neighbors, cell_labels=self.adata.obs_names)
        else:
            neighbors, weights = neighbors_and_weights_from_distances(
                self.distances, cell_index=self.adata.obs_names, n_neighbors=n_neighbors,
                neighborhood_factor=neighborhood_factor)

        neighbors = neighbors.loc[self.adata.obs_names]
        weights = weights.loc[self.adata.obs_names]

        self.neighbors = neighbors

        if not weighted_graph:
            weights = pd.DataFrame(
                np.ones_like(weights.values),
                index=weights.index, columns=weights.columns
            )

        weights = make_weights_non_redundant(neighbors.values, weights.values)

        weights = pd.DataFrame(
            weights, index=neighbors.index, columns=neighbors.columns)

        self.weights = weights

    def _compute_hotspot(self, jobs=1):
        """Perform feature selection using local autocorrelation

        In addition to returning output, this also stores the output
        in `self.results`.

        Alias for `self.compute_autocorrelations`

        Parameters
        ----------
        jobs: int
            Number of parallel jobs to run

        Returns
        -------
        results : pandas.DataFrame
            A dataframe with four columns:

              - C: Scaled -1:1 autocorrelation coeficients
              - Z: Z-score for autocorrelation
              - Pval:  P-values computed from Z-scores
              - FDR:  Q-values using the Benjamini-Hochberg procedure

            Gene ids are in the index

        """

        results = compute_hs(
            self.counts, self.neighbors, self.weights,
            self.umi_counts, self.model, genes=self.adata.var_names,
            centered=True, jobs=jobs)

        self.results = results

        return self.results

    def compute_autocorrelations(self, jobs=1):
        """Perform feature selection using local autocorrelation

        In addition to returning output, this also stores the output
        in `self.results`

        Parameters
        ----------
        jobs: int
            Number of parallel jobs to run

        Returns
        -------
        results : pandas.DataFrame
            A dataframe with four columns:

              - C: Scaled -1:1 autocorrelation coeficients
              - Z: Z-score for autocorrelation
              - Pval:  P-values computed from Z-scores
              - FDR:  Q-values using the Benjamini-Hochberg procedure

            Gene ids are in the index

        """
        return self._compute_hotspot(jobs)

    def compute_local_correlations(self, genes, jobs=1):
        """Define gene-gene relationships with pair-wise local correlations

        In addition to returning output, this method stores its result
        in `self.local_correlation_z`

        Parameters
        ----------
        genes: iterable of str
            gene identifies to compute local correlations on
            should be a smaller subset of all genes
        jobs: int
            Number of parallel jobs to run

        Returns
        -------
        local_correlation_z : pd.Dataframe
                local correlation Z-scores between genes
                shape is genes x genes
        """

        print(
            "Computing pair-wise local correlation on {} features..."
            .format(len(genes))
        )
        counts_dense = self._counts_from_anndata(
            self.adata[:, genes],
            self.layer_key,
            dense=True,
            pandas=True,
        )

        lc, lcz = compute_hs_pairs_centered_cond(
            counts_dense, self.neighbors, self.weights,
            self.umi_counts, self.model, jobs=jobs)

        self.local_correlation_c = lc
        self.local_correlation_z = lcz

        return self.local_correlation_z

    def create_modules(
            self, min_gene_threshold=20, core_only=True, fdr_threshold=0.05
    ):
        """Groups genes into modules

        In addition to being returned, the results of this method are retained
        in the object at `self.modules`.  Additionally, the linkage matrix
        (in the same form as that of scipy.cluster.hierarchy.linkage) is saved
        in `self.linkage` for plotting or manual clustering.

        Parameters
        ----------
        min_gene_threshold: int
            Controls how small modules can be.  Increase if there are too many
            modules being formed.  Decrease if substructre is not being captured
        core_only: bool
            Whether or not to assign ambiguous genes to a module or leave unassigned
        fdr_threshold: float
            Correlation theshold at which to stop assigning genes to modules

        Returns
        -------
        modules: pandas.Series
            Maps gene to module number.  Unassigned genes are indicated with -1


        """

        gene_modules, Z = modules.compute_modules(
            self.local_correlation_z, min_gene_threshold=min_gene_threshold,
            fdr_threshold=fdr_threshold, core_only=core_only
        )

        self.modules = gene_modules
        self.linkage = Z

        return self.modules

    def calculate_module_scores(self):
        """Calculate Module Scores

        In addition to returning its result, this method stores
        its output in the object at `self.module_scores`

        Returns
        -------
        module_scores: pandas.DataFrame
            Scores for each module for each gene
            Dimensions are genes x modules

        """

        modules_to_compute = sorted(
            [x for x in self.modules.unique() if x != -1]
        )

        print(
            "Computing scores for {} modules..."
            .format(len(modules_to_compute))
        )

        module_scores = {}
        for module in tqdm(modules_to_compute):
            module_genes = self.modules.index[self.modules == module]

            counts_dense = self._counts_from_anndata(
                self.adata[:, module_genes],
                self.layer_key,
                dense=True
            )

            scores = modules.compute_scores(
                counts_dense,
                self.model, self.umi_counts.values,
                self.neighbors.values, self.weights.values
            )

            module_scores[module] = scores

        module_scores = pd.DataFrame(module_scores)
        module_scores.index = self.adata.obs_names

        self.module_scores = module_scores

        return self.module_scores

    def plot_local_correlations(
            self, mod_cmap='tab10', vmin=-8, vmax=8,
            z_cmap='RdBu_r', yticklabels=False
    ):
        """Plots a clustergrid of the local correlation values

        Parameters
        ----------
        mod_cmap: valid matplotlib colormap str or object
            discrete colormap for module assignments on the left side
        vmin: float
            minimum value for colorscale for Z-scores
        vmax: float
            maximum value for colorscale for Z-scores
        z_cmap: valid matplotlib colormap str or object
            continuous colormap for correlation Z-scores
        yticklabels: bool
            Whether or not to plot all gene labels
            Default is false as there are too many.  However
            if using this plot interactively you may with to set
            to true so you can zoom in and read gene names
        """

        return local_correlation_plot(
                    self.local_correlation_z, self.modules, self.linkage,
                    mod_cmap=mod_cmap, vmin=vmin, vmax=vmax,
                    z_cmap=z_cmap, yticklabels=yticklabels
        )



name = "hotspot"

from .hotspot import Hotspot

# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
# https://github.com/python-poetry/poetry/issues/144#issuecomment-623927302
try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata
package_name = "hotspotsc"
__version__ = importlib_metadata.version(package_name)




def fit_gene_model(gene_counts, umi_counts):

    tj = gene_counts.sum()
    tis = umi_counts
    total = tis.sum()

    N = gene_counts.size

    min_size = 10**(-10)

    mu = tj*tis/total
    vv = (gene_counts - mu).var()*(N/(N-1))
    my_rowvar = vv

    # size = (tj**2) * (tis**2).sum()/total**2 / ((N-1)*my_rowvar-tj)
    # regroup terms to protect against overflow errors
    size = ((tj**2) / total) * ((tis**2).sum() / total) / ((N-1)*my_rowvar-tj)

    if size < 0:    # Can't have negative dispersion
        size = 1e9

    if size < min_size and size >= 0:
        size = min_size

    var = mu*(1+mu/size)
    x2 = var+mu**2

    return mu, var, x2



import numpy as np
from numba import jit
from tqdm import tqdm


def sim_latent(N_CELLS, N_DIM):
    """
    Draws a latent space coordinate for each cell
    randomly from a unit gaussian
    """

    return np.random.normal(size=(N_CELLS, N_DIM))


def sim_umi_counts(N_CELLS, mid, scale):
    """
    Simulates counts from a log-normal distribution
    """

    mu = np.log(mid)
    sd = np.log(mid+scale)-mu

    vals = np.random.normal(loc=mu, scale=sd, size=N_CELLS)
    vals = np.floor(np.exp(vals))

    return vals


def sim_counts_bernoulli(N_CELLS, umi_counts, gene_p):
    """
    For a given transcript probability, simulates detection
    in N_CELLS with `umi_counts` umis per cell

    Note: gene_p is in transcripts / 10,000
    """
    gene_p = gene_p / 10000

    detect_p = 1-(1-gene_p)**umi_counts

    vals = (np.random.rand(N_CELLS) < detect_p).astype('double')

    return vals


def sim_counts_danb(N_CELLS, N_GENES, umi_counts, mean, size):
    """
    Simulates transcript counts under the DANB model
    Uses umi_counts to adjust the mean

    Returns N_GENES replicates -> result is N_GENES x N_CELLS
    """
    umi_scale = umi_counts / umi_counts.mean()

    vals = np.zeros((N_GENES, N_CELLS))

    for i in tqdm(range(N_CELLS)):
        mean_i = mean*umi_scale[i]
        var = mean_i*(1+mean_i/size)
        p = 1 - mean_i/var
        n = mean_i * (1-p) / p
        p = 1-p

        vals[:, i] = np.random.negative_binomial(n=n, p=p, size=N_GENES)

    return vals


def generate_permutation_null(observed_counts, N_REPS):

    out = []
    for i in range(N_REPS):
        out.append(np.random.permutation(observed_counts))

    return np.vstack(out)


def generate_resampled_null(observed_counts, N_REPS):

    out = []
    for i in range(N_REPS):
        out.append(
            np.random.choice(
                observed_counts, size=len(observed_counts), replace=True
            )
        )

    return np.vstack(out)



"""
Simplest Model - just assumes expression data is normal
UMI counts are regressed out
"""
import numpy as np


def fit_gene_model(gene_counts, umi_counts):

    X = np.vstack((np.ones(len(umi_counts)), umi_counts)).T
    y = gene_counts.reshape((-1, 1))

    if umi_counts.var() == 0:

        mu = gene_counts.mean()
        var = gene_counts.var()
        mu = np.repeat(mu, len(umi_counts))
        var = np.repeat(var, len(umi_counts))
        x2 = mu**2 + var

        return mu, var, x2

    B = np.linalg.inv(X.T.dot(X)).dot(X.T).dot(y)

    mu = X.dot(B)

    var = (y - mu).var()
    var = np.repeat(var, len(umi_counts))

    mu = mu.ravel()

    x2 = mu**2 + var

    return mu, var, x2



import numpy as np
from numba import jit
from tqdm import tqdm
import pandas as pd
from scipy.stats import norm
from scipy.sparse import issparse
from statsmodels.stats.multitest import multipletests
import multiprocessing

from . import danb_model
from . import bernoulli_model
from . import normal_model
from . import none_model

from .knn import compute_node_degree
from .utils import center_values


@jit(nopython=True)
def local_cov_weights(x, neighbors, weights):
    out = 0

    for i in range(len(x)):
        for k in range(neighbors.shape[1]):

            j = neighbors[i, k]
            w_ij = weights[i, k]

            xi = x[i]
            xj = x[j]
            if xi == 0 or xj == 0 or w_ij == 0:
                out += 0
            else:
                out += xi * xj * w_ij

    return out


@jit(nopython=True)
def compute_moments_weights_slow(mu, x2, neighbors, weights):
    """
    This version exaustively iterates over all |E|^2 terms
    to compute the expected moments exactly.  Used to test
    the more optimized formulations that follow
    """

    N = neighbors.shape[0]
    K = neighbors.shape[1]

    # Calculate E[G]
    EG = 0
    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]

            EG += wij * mu[i] * mu[j]

    # Calculate E[G^2]
    EG2 = 0
    for i in range(N):

        EG2_i = 0

        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]

            for x in range(N):
                for z in range(K):
                    y = neighbors[x, z]
                    wxy = weights[x, z]

                    s = wij * wxy
                    if s == 0:
                        continue

                    if i == x:
                        if j == y:
                            t1 = x2[i] * x2[j]
                        else:
                            t1 = x2[i] * mu[j] * mu[y]
                    elif i == y:
                        if j == x:
                            t1 = x2[i] * x2[j]
                        else:
                            t1 = x2[i] * mu[j] * mu[x]
                    else:  # i is unique since i can't equal j

                        if j == x:
                            t1 = mu[i] * x2[j] * mu[y]
                        elif j == y:
                            t1 = mu[i] * x2[j] * mu[x]
                        else:  # i and j are unique, no shared nodes
                            t1 = mu[i] * mu[j] * mu[x] * mu[y]

                    EG2_i += s * t1

        EG2 += EG2_i

    return EG, EG2


@jit(nopython=True)
def compute_moments_weights(mu, x2, neighbors, weights):

    N = neighbors.shape[0]
    K = neighbors.shape[1]

    # Calculate E[G]
    EG = 0
    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]

            EG += wij * mu[i] * mu[j]

    # Calculate E[G^2]
    EG2 = 0

    #   Get the x^2*y*z terms
    t1 = np.zeros(N)
    t2 = np.zeros(N)

    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]

            wij = weights[i, k]
            if wij == 0:
                continue

            t1[i] += wij * mu[j]
            t2[i] += wij**2 * mu[j] ** 2

            t1[j] += wij * mu[i]
            t2[j] += wij**2 * mu[i] ** 2

    t1 = t1**2

    for i in range(N):
        EG2 += (x2[i] - mu[i] ** 2) * (t1[i] - t2[i])

    #  Get the x^2*y^2 terms
    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]

            wij = weights[i, k]

            EG2 += wij**2 * (x2[i] * x2[j] - (mu[i] ** 2) * (mu[j] ** 2))

    EG2 += EG**2

    return EG, EG2


@jit(nopython=True)
def compute_local_cov_max(node_degrees, vals):
    tot = 0.0

    for i in range(node_degrees.size):
        tot += node_degrees[i] * (vals[i] ** 2)

    return tot / 2


def compute_hs(
    counts, neighbors, weights, num_umi, model, genes, centered=False, jobs=1
):

    neighbors = neighbors.values
    weights = weights.values
    num_umi = num_umi.values

    D = compute_node_degree(neighbors, weights)
    Wtot2 = (weights**2).sum()

    def data_iter():
        for i in range(counts.shape[0]):
            vals = counts[i]
            if issparse(vals):
                vals = np.asarray(vals.A).ravel()
            vals = vals.astype("double")
            yield vals

    def initializer():
        global g_neighbors
        global g_weights
        global g_num_umi
        global g_model
        global g_centered
        global g_Wtot2
        global g_D
        g_neighbors = neighbors
        g_weights = weights
        g_num_umi = num_umi
        g_model = model
        g_centered = centered
        g_Wtot2 = Wtot2
        g_D = D

    if jobs > 1:

        with multiprocessing.Pool(processes=jobs, initializer=initializer) as pool:

            results = list(
                tqdm(pool.imap(_map_fun_parallel, data_iter()), total=counts.shape[0])
            )
    else:

        def _map_fun(vals):
            return _compute_hs_inner(
                vals, neighbors, weights, num_umi, model, centered, Wtot2, D
            )

        results = list(tqdm(map(_map_fun, data_iter()), total=counts.shape[0]))

    results = pd.DataFrame(results, index=genes, columns=["G", "EG", "stdG", "Z", "C"])

    results["Pval"] = norm.sf(results["Z"].values)
    results["FDR"] = multipletests(results["Pval"], method="fdr_bh")[1]

    results = results.sort_values("Z", ascending=False)
    results.index.name = "Gene"

    results = results[["C", "Z", "Pval", "FDR"]]  # Remove other columns

    return results


def _compute_hs_inner(vals, neighbors, weights, num_umi, model, centered, Wtot2, D):
    """
    Note, since this is an inner function, for parallelization to work well
    none of the contents of the function can use MKL or OPENBLAS threads.
    Or else we open too many.  Because of this, some simple numpy operations
    are re-implemented using numba instead as it's difficult to control
    the number of threads in numpy after it's imported
    """

    if model == "bernoulli":
        vals = (vals > 0).astype("double")
        mu, var, x2 = bernoulli_model.fit_gene_model(vals, num_umi)
    elif model == "danb":
        mu, var, x2 = danb_model.fit_gene_model(vals, num_umi)
    elif model == "normal":
        mu, var, x2 = normal_model.fit_gene_model(vals, num_umi)
    elif model == "none":
        mu, var, x2 = none_model.fit_gene_model(vals, num_umi)
    else:
        raise Exception("Invalid Model: {}".format(model))

    if centered:
        vals = center_values(vals, mu, var)

    G = local_cov_weights(vals, neighbors, weights)

    if centered:
        EG, EG2 = 0, Wtot2
    else:
        EG, EG2 = compute_moments_weights(mu, x2, neighbors, weights)

    stdG = (EG2 - EG * EG) ** 0.5

    Z = (G - EG) / stdG

    G_max = compute_local_cov_max(D, vals)
    C = (G - EG) / G_max

    return [G, EG, stdG, Z, C]


def _map_fun_parallel(vals):
    global g_neighbors
    global g_weights
    global g_num_umi
    global g_model
    global g_centered
    global g_Wtot2
    global g_D
    return _compute_hs_inner(
        vals, g_neighbors, g_weights, g_num_umi, g_model, g_centered, g_Wtot2, g_D
    )



from numba import jit
import pandas as pd
import numpy as np
from numba import njit

N_BIN_TARGET = 30


@jit(nopython=True)
def find_gene_p(num_umi, D):
    """
    Finds gene_p such that sum of expected detects
    matches our data

    Performs a binary search on p in the space of log(p)
    """

    low = 1e-12
    high = 1

    if D == 0:
        return 0

    for ITER in range(40):

        attempt = (high*low)**0.5
        tot = 0

        for i in range(len(num_umi)):
            tot = tot + 1-(1-attempt)**num_umi[i]

        if abs(tot-D)/D < 1e-3:
            break

        if tot > D:
            high = attempt
        else:
            low = attempt

    return (high*low)**0.5


def fit_gene_model_scaled(gene_detects, umi_counts):

    D = gene_detects.sum()

    gene_p = find_gene_p(umi_counts, D)

    detect_p = 1-(1-gene_p)**umi_counts

    mu = detect_p
    var = detect_p * (1 - detect_p)
    x2 = detect_p

    return mu, var, x2


def true_params_scaled(gene_p, umi_counts):

    detect_p = 1-(1-gene_p/10000)**umi_counts

    mu = detect_p
    var = detect_p * (1 - detect_p)
    x2 = detect_p

    return mu, var, x2


def fit_gene_model_linear(gene_detects, umi_counts):

    umi_count_bins, bins = pd.qcut(
        np.log10(umi_counts), N_BIN_TARGET, labels=False, retbins=True,
        duplicates='drop'
    )
    bin_centers = np.array(
        [bins[i] / 2 + bins[i + 1] / 2 for i in range(len(bins) - 1)]
    )

    N_BIN = len(bin_centers)

    bin_detects = bin_gene_detection(gene_detects, umi_count_bins, N_BIN)

    lbin_detects = logit(bin_detects)

    X = np.ones((N_BIN, 2))
    X[:, 1] = bin_centers
    Y = lbin_detects

    b = np.linalg.inv(X.T.dot(X)).dot(X.T).dot(Y)
    detect_p = ilogit(b[0] + b[1] * np.log10(umi_counts))

    mu = detect_p
    var = detect_p * (1 - detect_p)
    x2 = detect_p

    return mu, var, x2


fit_gene_model = fit_gene_model_linear


@njit
def logit(p):
    return np.log(p / (1 - p))


@njit
def ilogit(q):
    return np.exp(q) / (1 + np.exp(q))


@njit
def bin_gene_detection(gene_detects, umi_count_bins, N_BIN):
    bin_detects = np.zeros(N_BIN)
    bin_totals = np.zeros(N_BIN)

    for i in range(len(gene_detects)):
        x = gene_detects[i]
        bin_i = umi_count_bins[i]
        bin_detects[bin_i] += x
        bin_totals[bin_i] += 1

    # Need to account for 0% detects
    #    Add 1 to numerator and denominator
    # Need to account for 100% detects
    #    Add 1 to denominator

    return (bin_detects+1) / (bin_totals+2)



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
from scipy.cluster.hierarchy import leaves_list


def local_correlation_plot(
            local_correlation_z, modules, linkage,
            mod_cmap='tab10', vmin=-8, vmax=8,
            z_cmap='RdBu_r', yticklabels=False
):

    row_colors = None
    colors = list(plt.get_cmap(mod_cmap).colors)
    module_colors = {i: colors[(i-1) % len(colors)] for i in modules.unique()}
    module_colors[-1] = '#ffffff'

    row_colors1 = pd.Series(
        [module_colors[i] for i in modules],
        index=local_correlation_z.index,
    )

    row_colors = pd.DataFrame({
        "Modules": row_colors1,
    })

    cm = sns.clustermap(
        local_correlation_z,
        row_linkage=linkage,
        col_linkage=linkage,
        vmin=vmin,
        vmax=vmax,
        cmap=z_cmap,
        xticklabels=False,
        yticklabels=yticklabels,
        row_colors=row_colors,
        rasterized=True,
    )

    fig = plt.gcf()
    plt.sca(cm.ax_heatmap)
    plt.ylabel("")
    plt.xlabel("")

    cm.ax_row_dendrogram.remove()

    # Add 'module X' annotations
    ii = leaves_list(linkage)

    mod_reordered = modules.iloc[ii]

    mod_map = {}
    y = np.arange(modules.size)

    for x in mod_reordered.unique():
        if x == -1:
            continue

        mod_map[x] = y[mod_reordered == x].mean()

    plt.sca(cm.ax_row_colors)
    for mod, mod_y in mod_map.items():
        plt.text(-.5, y=mod_y, s="Module {}".format(mod),
                 horizontalalignment='right',
                 verticalalignment='center')
    plt.xticks([])

    # Find the colorbar 'child' and modify
    min_delta = 1e99
    min_aa = None
    for aa in fig.get_children():
        try:
            bbox = aa.get_position()
            delta = (0-bbox.xmin)**2 + (1-bbox.ymax)**2
            if delta < min_delta:
                delta = min_delta
                min_aa = aa
        except AttributeError:
            pass

    min_aa.set_ylabel('Z-Scores')
    min_aa.yaxis.set_label_position("left")



import numpy as np
import pandas as pd

from sklearn.decomposition import PCA
from .utils import neighbor_smoothing_row
from .local_stats_pairs import create_centered_counts_row
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage
from scipy.stats import norm
from statsmodels.stats.multitest import multipletests


def compute_scores(
        counts_sub, model, num_umi, neighbors, weights):
    """
    counts_sub: row-subset of counts matrix with genes in the module
    """

    cc_smooth = np.zeros_like(counts_sub, dtype=np.float64)

    for i in range(counts_sub.shape[0]):

        counts_row = counts_sub[i, :]
        centered_row = create_centered_counts_row(counts_row, model, num_umi)
        smooth_row = neighbor_smoothing_row(
            centered_row, neighbors, weights, _lambda=.9)

        cc_smooth[i] = smooth_row

    pca_data = cc_smooth

    model = PCA(n_components=1)
    scores = model.fit_transform(pca_data.T)

    sign = model.components_.mean()  # may need to flip
    if sign < 0:
        scores = scores * -1

    scores = scores[:, 0]

    return scores


def sort_linkage(Z, node_index, node_values):
    """
    Sorts linkage by 'node_values' in place
    """

    N = Z.shape[0] + 1  # number of leaves

    if node_index < 0:
        return

    left_child = int(Z[node_index, 0] - N)
    right_child = int(Z[node_index, 1] - N)

    swap = False

    if left_child < 0 and right_child < 0:
        swap = False
    elif left_child < 0 and right_child >= 0:
        swap = True
    elif left_child >= 0 and right_child < 0:
        swap = False
    else:
        if node_values[left_child] > node_values[right_child]:
            swap = True
        else:
            swap = False

    if swap:
        Z[node_index, 0] = right_child + N
        Z[node_index, 1] = left_child + N

    sort_linkage(Z, left_child, node_values)
    sort_linkage(Z, right_child, node_values)


def calc_mean_dists(Z, node_index, out_mean_dists):
    """
    Calculates the mean density of joins
    for sub-trees underneath each node
    """

    N = Z.shape[0] + 1  # number of leaves

    left_child = int(Z[node_index, 0] - N)
    right_child = int(Z[node_index, 1] - N)

    if left_child < 0:
        left_average = 0
        left_merges = 0
    else:
        left_average, left_merges = calc_mean_dists(
            Z, left_child, out_mean_dists
        )

    if right_child < 0:
        right_average = 0
        right_merges = 0
    else:
        right_average, right_merges = calc_mean_dists(
            Z, right_child, out_mean_dists
        )

    this_height = Z[node_index, 2]
    this_merges = left_merges + right_merges + 1
    this_average = (
        left_average * left_merges + right_average * right_merges + this_height
    ) / this_merges

    out_mean_dists[node_index] = this_average

    return this_average, this_merges


def prop_label(Z, node_index, label, labels, out_clusters):
    """
    Propagates node labels downward if they are not -1
    Used to find the correct cluster label at the leaves
    """

    N = Z.shape[0] + 1  # number of leaves

    if label == -1:
        label = labels[node_index]

    left_child = int(Z[node_index, 0] - N)
    right_child = int(Z[node_index, 1] - N)

    if left_child < 0:
        out_clusters[left_child + N] = label
    else:
        prop_label(Z, left_child, label, labels, out_clusters)

    if right_child < 0:
        out_clusters[right_child + N] = label
    else:
        prop_label(Z, right_child, label, labels, out_clusters)


def prop_label2(Z, node_index, label, labels, out_clusters):
    """
    Propagates node labels downward
    Helper method used in assign_modules
    """

    N = Z.shape[0] + 1  # number of leaves

    parent_label = label
    this_label = labels[node_index]

    if this_label == -1:
        new_label = parent_label
    else:
        new_label = this_label

    left_child = int(Z[node_index, 0] - N)
    right_child = int(Z[node_index, 1] - N)

    if left_child < 0:
        out_clusters[left_child + N] = new_label
    else:
        prop_label2(Z, left_child, new_label, labels, out_clusters)

    if right_child < 0:
        out_clusters[right_child + N] = new_label
    else:
        prop_label2(Z, right_child, new_label, labels, out_clusters)


def assign_modules(Z, leaf_labels, offset, MIN_THRESHOLD=10, Z_THRESHOLD=3):
    clust_i = 0

    labels = np.ones(Z.shape[0])*-1
    N = Z.shape[0]+1

    mean_dists = np.zeros(Z.shape[0])
    calc_mean_dists(Z, Z.shape[0]-1, mean_dists)

    for i in range(Z.shape[0]):

        ca = int(Z[i, 0])
        cb = int(Z[i, 1])

        if ca - N < 0:  # leaf node
            n_members_a = 1
            clust_a = -1
        else:
            n_members_a = Z[ca-N, 3]
            clust_a = labels[ca-N]

        if cb - N < 0:  # leaf node
            n_members_b = 1
            clust_b = -1
        else:
            n_members_b = Z[cb-N, 3]
            clust_b = labels[cb-N]

        if Z[i, 2] > offset - Z_THRESHOLD:
            new_clust_assign = -1
        elif (n_members_a >= MIN_THRESHOLD and n_members_b >= MIN_THRESHOLD):
            # don't join them
            # assign the one with the larger mean distance
            dist_a = mean_dists[ca-N]
            dist_b = mean_dists[cb-N]
            if dist_a >= dist_b:
                new_clust_assign = clust_a
            else:
                new_clust_assign = clust_b
        elif n_members_a >= MIN_THRESHOLD:
            new_clust_assign = clust_a
        elif n_members_b >= MIN_THRESHOLD:
            new_clust_assign = clust_b
        elif (n_members_b + n_members_a) >= MIN_THRESHOLD:
            # A new cluster is born!
            new_clust_assign = clust_i
            clust_i += 1
        else:
            new_clust_assign = -1  # Still too small

        labels[i] = new_clust_assign

    out_clusters = np.ones(N)*-2
    prop_label2(Z, Z.shape[0]-1, labels[-1], labels, out_clusters)

    # remap out_clusters
    unique_clusters = list(np.sort(np.unique(out_clusters)))

    if -1 in unique_clusters:
        unique_clusters.remove(-1)

    clust_map = {
        x: i+1 for i, x in enumerate(unique_clusters)
    }
    clust_map[-1] = -1

    out_clusters = [clust_map[x] for x in out_clusters]
    out_clusters = pd.Series(out_clusters, index=leaf_labels)

    return out_clusters


def assign_modules_core(Z, leaf_labels, offset, MIN_THRESHOLD=10, Z_THRESHOLD=3):
    clust_i = 0

    labels = np.ones(Z.shape[0])*-1
    N = Z.shape[0]+1

    for i in range(Z.shape[0]):

        ca = int(Z[i, 0])
        cb = int(Z[i, 1])

        if ca - N < 0:  # leaf node
            n_members_a = 1
            clust_a = -1
        else:
            n_members_a = Z[ca-N, 3]
            clust_a = labels[ca-N]

        if cb - N < 0:  # leaf node
            n_members_b = 1
            clust_b = -1
        else:
            n_members_b = Z[cb-N, 3]
            clust_b = labels[cb-N]

        if (n_members_a >= MIN_THRESHOLD and n_members_b >= MIN_THRESHOLD):
            # don't join them
            new_clust_assign = -1
        elif Z[i, 2] > offset - Z_THRESHOLD:
            new_clust_assign = -1
        elif n_members_a >= MIN_THRESHOLD:
            new_clust_assign = clust_a
        elif n_members_b >= MIN_THRESHOLD:
            new_clust_assign = clust_b
        elif (n_members_b + n_members_a) >= MIN_THRESHOLD:
            # A new cluster is born!
            new_clust_assign = clust_i
            clust_i += 1
        else:
            new_clust_assign = -1  # Still too small

        labels[i] = new_clust_assign

    out_clusters = np.ones(N)*-2
    prop_label(Z, Z.shape[0]-1, labels[-1], labels, out_clusters)

    # remap out_clusters
    unique_clusters = list(np.sort(np.unique(out_clusters)))

    if -1 in unique_clusters:
        unique_clusters.remove(-1)

    clust_map = {
        x: i+1 for i, x in enumerate(unique_clusters)
    }
    clust_map[-1] = -1

    out_clusters = [clust_map[x] for x in out_clusters]
    out_clusters = pd.Series(out_clusters, index=leaf_labels)

    return out_clusters


def compute_modules(Z_scores, min_gene_threshold=10, fdr_threshold=None, z_threshold=None, core_only=False):
    """
    Assigns modules from the gene pair-wise Z-scores

    Parameters
    ----------
    Z_scores: pandas.DataFrame
        local correlations between genes
    min_gene_threshold: int, optional
        minimum number of genes to create a module
    fdr_threshold: float, optional
        used to determine minimally significant z_score
    core_only: bool, optional
        whether or not to assign unassigned genes to a module

    Returns
    -------
    modules: pandas.Series
        maps gene id to module id
    linkage: numpy.ndarray
        Linkage matrix in the format used by scipy.cluster.hierarchy.linkage

    """

    # Determine Z_Threshold from FDR threshold

    if z_threshold is None:
        allZ = squareform(  # just in case slightly not symmetric
            Z_scores.values/2 + Z_scores.values.T/2
        )
        allZ = np.sort(allZ)
        allP = norm.sf(allZ)
        allP_c = multipletests(allP, method='fdr_bh')[1]
        ii = np.nonzero(allP_c < fdr_threshold)[0]
        if ii.size > 0:
            z_threshold = allZ[ii[0]]
        else:
            z_threshold = allZ[-1]+1

    # Compute the linkage matrix
    dd = Z_scores.copy().values
    np.fill_diagonal(dd, 0)
    condensed = squareform(dd)*-1
    offset = condensed.min() * -1
    condensed += offset
    Z = linkage(condensed, method='average')

    # Linkage -> Modules
    if core_only:
        out_clusters = assign_modules_core(
            Z, offset=offset, MIN_THRESHOLD=min_gene_threshold,
            leaf_labels=Z_scores.index, Z_THRESHOLD=z_threshold)
    else:
        out_clusters = assign_modules(
            Z, offset=offset, MIN_THRESHOLD=min_gene_threshold,
            leaf_labels=Z_scores.index, Z_THRESHOLD=z_threshold)

    # Sort the leaves of the linkage matrix (for plotting)
    mean_dists = np.zeros(Z.shape[0])
    calc_mean_dists(Z, Z.shape[0]-1, mean_dists)
    linkage_out = Z.copy()
    sort_linkage(linkage_out, Z.shape[0]-1, mean_dists)

    out_clusters.name = 'Module'

    return out_clusters, linkage_out



import numpy as np
from numba import jit, njit
from tqdm import tqdm
import pandas as pd
import multiprocessing
import itertools

from . import danb_model
from . import bernoulli_model
from . import normal_model
from . import none_model
from .local_stats import compute_local_cov_max
from .knn import compute_node_degree
from .utils import center_values


@jit(nopython=True)
def conditional_eg2(x, neighbors, weights):
    """
    Computes EG2 for the conditional correlation
    """
    N = neighbors.shape[0]
    K = neighbors.shape[1]

    t1x = np.zeros(N)

    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]

            wij = weights[i, k]
            if wij == 0:
                continue

            t1x[i] += wij*x[j]
            t1x[j] += wij*x[i]

    out_eg2 = (t1x**2).sum()

    return out_eg2


def conditional_eg2_slow(x, neighbors, weights):
    """
    Computes EG2 for the conditional correlation
    This version is slower and more explicit for debugging purposes
    """
    N = neighbors.shape[0]
    K = neighbors.shape[1]

    # enumerate ALL neighbors of each node and the edge weight

    neighbors_all = [[] for _ in range(N)]
    weights_all = [[] for _ in range(N)]

    for i in range(N):
        neighbors_i = neighbors_all[i]
        weights_i = weights_all[i]

        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]
            if wij == 0:
                continue

            neighbors_j = neighbors_all[j]
            weights_j = weights_all[j]

            neighbors_i.append(j)
            weights_i.append(wij)

            neighbors_j.append(i)
            weights_j.append(wij)

    node_totals = np.zeros(N)
    node_totals_diag = np.zeros(N)

    for i in range(N):
        neighbors_i = neighbors_all[i]
        weights_i = weights_all[i]

        for j, wij in zip(neighbors_i, weights_i):
            node_totals[i] += x[j]*wij
            node_totals_diag[i] += (x[j]*wij)**2

    node_totals = node_totals**2

    dup = 0

    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]
            if wij == 0:
                continue

            dup += wij**2*(x[i]**2+x[j]**2)

    return node_totals.sum() - node_totals_diag.sum() + dup


@jit(nopython=True)
def local_cov_pair(x, y, neighbors, weights):
    """Test statistic for local pair-wise autocorrelation"""
    out = 0

    for i in range(len(x)):
        xi = x[i]
        yi = y[i]
        if xi == 0 and yi == 0:
            continue
        for k in range(neighbors.shape[1]):

            j = neighbors[i, k]
            w_ij = weights[i, k]

            xj = x[j]
            yj = y[j]

            out += w_ij*(xi*yj + yi*xj)/2

    return out


@jit(nopython=True)
def compute_moments_weights_pairs_slow(muA, a2, muB, b2, neighbors, weights):
    """
    This version exaustively iterates over all |E|^2 terms
    to compute the expected moments exactly.  Used to test
    the more optimized formulations that follow
    """

    N = neighbors.shape[0]
    K = neighbors.shape[1]

    # Calculate E[G]
    EG = 0
    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]

            EG += wij * (muA[i] * muB[j] + muB[i] * muA[j]) / 2

    # Calculate E[G^2]
    EG2 = 0
    for i in range(N):

        EG2_i = 0

        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]

            for x in range(N):
                for z in range(K):
                    y = neighbors[x, z]
                    wxy = weights[x, z]

                    s = wij * wxy
                    if s == 0:
                        continue

                    if i == x:
                        if j == y:
                            t1 = (
                                a2[i] * b2[j]
                                + muA[i] * muB[j] * muA[j] * muB[i]
                                + muA[j] * muB[i] * muA[i] * muB[j]
                                + a2[j] * b2[i]
                            )
                        else:
                            t1 = (
                                a2[i] * muB[j] * muB[y]
                                + muA[i] * muB[j] * muA[y] * muB[i]
                                + muA[j] * muB[i] * muA[i] * muB[y]
                                + muA[j] * b2[i] * muA[y]
                            )
                    elif i == y:
                        if j == x:
                            t1 = (
                                muA[i] * muB[j] * muA[j] * muB[i]
                                + a2[i] * b2[j]
                                + a2[j] * b2[i]
                                + muA[j] * muB[i] * muA[i] * muB[j]
                            )
                        else:
                            t1 = (
                                muA[i] * muB[j] * muA[x] * muB[i]
                                + a2[i] * muB[j] * muB[x]
                                + muA[j] * b2[i] * muA[x]
                                + muA[j] * muB[i] * muA[i] * muB[x]
                            )
                    else:  # i is unique since i can't equal j

                        if j == x:
                            t1 = (
                                muA[i] * muB[j] * muA[j] * muB[y]
                                + muA[i] * b2[j] * muA[y]
                                + a2[j] * muB[i] * muB[y]
                                + muA[j] * muB[i] * muA[y] * muB[j]
                            )
                        elif j == y:
                            t1 = (
                                muA[i] * b2[j] * muA[x]
                                + muA[i] * muB[j] * muA[j] * muB[x]
                                + muA[j] * muB[i] * muA[x] * muB[j]
                                + a2[j] * muB[i] * muB[x]
                            )
                        else:  # i and j are unique, no shared nodes
                            t1 = (
                                muA[i] * muB[j] * muA[x] * muB[y]
                                + muA[i] * muB[j] * muA[y] * muB[x]
                                + muA[j] * muB[i] * muA[x] * muB[y]
                                + muA[j] * muB[i] * muA[y] * muB[x]
                            )

                    EG2_i += s * t1 / 4

        EG2 += EG2_i

    return EG, EG2


@jit(nopython=True)
def compute_moments_weights_pairs(
        muX, x2,
        muY, y2,
        neighbors, weights):
    """Computes the expectations of the local pair-wise test statistic"""

    N = neighbors.shape[0]
    K = neighbors.shape[1]

    # Calculate E[G]
    EG = 0
    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]

            EG += wij*(muX[i]*muY[j] + muY[i]*muX[j])/2

    # Calculate E[G^2]
    EG2 = 0
    EG2 += (EG**2)

    #   Get the x^2*y*z terms
    t1x = np.zeros(N)
    t2x = np.zeros(N)

    t1y = np.zeros(N)
    t2y = np.zeros(N)

    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]

            wij = weights[i, k]
            if wij == 0:
                continue

            t1x[i] += wij*muX[j]
            t2x[i] += wij**2*muX[j]**2

            t1x[j] += wij*muX[i]
            t2x[j] += wij**2*muX[i]**2

            t1y[i] += wij*muY[j]
            t2y[i] += wij**2*muY[j]**2

            t1y[j] += wij*muY[i]
            t2y[j] += wij**2*muY[i]**2

    t1x = t1x**2
    t1y = t1y**2

    for i in range(N):
        EG2 += (y2[i] - muY[i]**2)*(t1x[i] - t2x[i])/4
        EG2 += (x2[i] - muX[i]**2)*(t1y[i] - t2y[i])/4

    #  Get the x^2*y^2 terms
    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]

            wij = weights[i, k]

            EG2 += wij**2*(
                x2[i]*y2[j] - (muX[i]**2)*(muY[j]**2) +
                y2[i]*x2[j] - (muY[i]**2)*(muX[j]**2)
            )/4

    return EG, EG2


@jit(nopython=True)
def compute_moments_weights_pairs_fast(
        muX, x2,
        muY, y2,
        neighbors, weights):
    """
    Computes the expectations of the local pair-wise test statistic

    About 2x as fast as compute_moments_weights_pairs
    """

    N = neighbors.shape[0]
    K = neighbors.shape[1]

    #   Get the x^2*y*z terms
    t1x = np.zeros(N)
    t2x = np.zeros(N)

    t1y = np.zeros(N)
    t2y = np.zeros(N)

    EG = 0
    EG2 = 0

    for i in range(N):
        for k in range(K):
            j = neighbors[i, k]
            wij = weights[i, k]
            wij_2 = wij**2

            if wij == 0:
                continue

            muX_i2 = muX[i]**2
            muX_j2 = muX[j]**2
            muY_i2 = muY[i]**2
            muY_j2 = muY[j]**2

            EG += wij*(muX[i]*muY[j] + muY[i]*muX[j])/2

            t1x[i] += wij*muX[j]
            t2x[i] += wij_2*muX_j2

            t1x[j] += wij*muX[i]
            t2x[j] += wij_2*muX_i2

            t1y[i] += wij*muY[j]
            t2y[i] += wij_2*muY_j2

            t1y[j] += wij*muY[i]
            t2y[j] += wij_2*muY_i2

            EG2 += wij_2*(
                x2[i]*y2[j] - (muX_i2)*(muY_j2) +
                y2[i]*x2[j] - (muY_i2)*(muX_j2)
            )/4

    # Calculate E[G^2]
    t1x = t1x**2
    t1y = t1y**2

    for i in range(N):
        EG2 += (y2[i] - muY[i]**2)*(t1x[i] - t2x[i])/4
        EG2 += (x2[i] - muX[i]**2)*(t1y[i] - t2y[i])/4

    EG2 += (EG**2)

    return EG, EG2


@jit(nopython=True)
def compute_moments_weights_pairs_std(neighbors, weights):
    """
    Computes the expectations of the local pair-wise test statistic

    This version assumes variables are standardized,
    and so the moments are actually the same for all
    pairs of variables
    """

    N = neighbors.shape[0]
    K = neighbors.shape[1]

    # Calculate E[G]
    EG = 0

    # Calculate E[G^2]
    EG2 = 0

    #  Get the x^2*y^2 terms
    for i in range(N):
        for k in range(K):
            wij = weights[i, k]

            EG2 += (wij**2)/2

    return EG, EG2


def create_centered_counts(counts, model, num_umi):
    """
    Creates a matrix of centered/standardized counts given
    the selected statistical model
    """
    out = np.zeros_like(counts, dtype='double')

    for i in tqdm(range(out.shape[0])):

        vals_x = counts[i]

        out_x = create_centered_counts_row(
            vals_x, model, num_umi)

        out[i] = out_x

    return out


def create_centered_counts_row(vals_x, model, num_umi):

    if model == 'bernoulli':
        vals_x = (vals_x > 0).astype('double')
        mu_x, var_x, x2_x = bernoulli_model.fit_gene_model(
            vals_x, num_umi)

    elif model == 'danb':
        mu_x, var_x, x2_x = danb_model.fit_gene_model(
            vals_x, num_umi)

    elif model == 'normal':
        mu_x, var_x, x2_x = normal_model.fit_gene_model(
            vals_x, num_umi)

    elif model == 'none':
        mu_x, var_x, x2_x = none_model.fit_gene_model(
            vals_x, num_umi)

    else:
        raise Exception("Invalid Model: {}".format(model))

    var_x[var_x == 0] = 1
    out_x = (vals_x-mu_x)/(var_x**0.5)
    out_x[out_x == 0] = 0

    return out_x


def _compute_hs_pairs_inner(row_i, counts, neighbors, weights, num_umi,
                            model, centered, Wtot2, D):

    vals_x = counts[row_i]

    lc_out = np.zeros(counts.shape[0])
    lc_z_out = np.zeros(counts.shape[0])

    if model == 'bernoulli':
        vals_x = (vals_x > 0).astype('double')
        mu_x, var_x, x2_x = bernoulli_model.fit_gene_model(
            vals_x, num_umi)

    elif model == 'danb':
        mu_x, var_x, x2_x = danb_model.fit_gene_model(
            vals_x, num_umi)

    elif model == 'normal':
        mu_x, var_x, x2_x = normal_model.fit_gene_model(
            vals_x, num_umi)

    elif model == 'none':
        mu_x, var_x, x2_x = none_model.fit_gene_model(
            vals_x, num_umi)

    else:
        raise Exception("Invalid Model: {}".format(model))

    if centered:
        vals_x = center_values(vals_x, mu_x, var_x)

    for row_j in range(counts.shape[0]):

        if row_j > row_i:
            continue

        vals_y = counts[row_j]

        if model == 'bernoulli':
            vals_y = (vals_y > 0).astype('double')
            mu_y, var_y, x2_y = bernoulli_model.fit_gene_model(
                vals_y, num_umi)

        elif model == 'danb':
            mu_y, var_y, x2_y = danb_model.fit_gene_model(
                vals_y, num_umi)

        elif model == 'normal':
            mu_y, var_y, x2_y = normal_model.fit_gene_model(
                vals_y, num_umi)

        elif model == 'none':
            mu_x, var_x, x2_x = none_model.fit_gene_model(
                vals_x, num_umi)

        else:
            raise Exception("Invalid Model: {}".format(model))

        if centered:
            vals_y = center_values(vals_y, mu_y, var_y)

        if centered:
            EG, EG2 = 0, Wtot2/2
        else:
            EG, EG2 = compute_moments_weights_pairs_fast(mu_x, x2_x,
                                                         mu_y, x2_y,
                                                         neighbors, weights)

        lc = local_cov_pair(vals_x, vals_y,
                            neighbors, weights)

        stdG = (EG2 - EG**2)**.5

        Z = (lc - EG) / stdG

        lc_out[row_j] = lc
        lc_z_out[row_j] = Z

    return (lc_out, lc_z_out)


def _compute_hs_pairs_inner_centered(
        rowpair, counts, neighbors, weights, Wtot2, D):
    """
    This version assumes that the counts have already been modeled
    and centered
    """
    row_i, row_j = rowpair

    vals_x = counts[row_i]
    vals_y = counts[row_j]

    EG, EG2 = 0, Wtot2/2

    lc = local_cov_pair(vals_x, vals_y,
                        neighbors, weights)

    stdG = (EG2 - EG**2)**.5

    Z = (lc - EG) / stdG

    return (lc, Z)


@jit(nopython=True)
def _compute_hs_pairs_inner_centered_cond_sym(
    rowpair, counts, neighbors, weights, eg2s
):
    """
    This version assumes that the counts have already been modeled
    and centered
    """
    row_i, row_j = rowpair

    vals_x = counts[row_i]
    vals_y = counts[row_j]

    lc = local_cov_pair(vals_x, vals_y, neighbors, weights)*2

    # Compute xy
    EG, EG2 = 0, eg2s[row_i]

    stdG = (EG2 - EG ** 2) ** 0.5

    Zxy = (lc - EG) / stdG

    # Compute yx
    EG, EG2 = 0, eg2s[row_j]

    stdG = (EG2 - EG ** 2) ** 0.5

    Zyx = (lc - EG) / stdG

    if abs(Zxy) < abs(Zyx):
        Z = Zxy
    else:
        Z = Zyx

    return (lc, Z)


def compute_hs_pairs(counts, neighbors, weights,
                     num_umi, model, centered=False, jobs=1):

    genes = counts.index

    counts = counts.values
    neighbors = neighbors.values
    weights = weights.values
    num_umi = num_umi.values

    D = compute_node_degree(neighbors, weights)
    Wtot2 = (weights**2).sum()

    def initializer():
        global g_neighbors
        global g_weights
        global g_num_umi
        global g_model
        global g_centered
        global g_Wtot2
        global g_D
        global g_counts
        g_counts = counts
        g_neighbors = neighbors
        g_weights = weights
        g_num_umi = num_umi
        g_model = model
        g_centered = centered
        g_Wtot2 = Wtot2
        g_D = D

    if jobs > 1:

        with multiprocessing.Pool(
                processes=jobs, initializer=initializer) as pool:

            results = list(
                tqdm(
                    pool.imap(_map_fun_parallel_pairs, range(counts.shape[0])),
                    total=counts.shape[0]
                )
            )
    else:
        def _map_fun(row_i):
            return _compute_hs_pairs_inner(
                row_i, counts, neighbors, weights, num_umi,
                model, centered, Wtot2, D)
        results = list(
            tqdm(
                map(_map_fun, range(counts.shape[0])),
                total=counts.shape[0]
            )
        )

    # Only have the lower triangle so we must rebuild the rest
    lcs = [x[0] for x in results]
    lc_zs = [x[1] for x in results]

    lcs = np.vstack(lcs)
    lc_zs = np.vstack(lc_zs)

    lcs = lcs + lcs.T
    lc_zs = lc_zs + lc_zs.T

    np.fill_diagonal(lcs, lcs.diagonal() / 2)
    np.fill_diagonal(lc_zs, lc_zs.diagonal() / 2)

    lc_maxs = compute_local_cov_pairs_max(D, counts)
    lcs = lcs / lc_maxs

    lcs = pd.DataFrame(lcs, index=genes, columns=genes)
    lc_zs = pd.DataFrame(lc_zs, index=genes, columns=genes)

    return lcs, lc_zs


def compute_hs_pairs_centered(counts, neighbors, weights,
                              num_umi, model, jobs=1):

    genes = counts.index

    counts = counts.values
    neighbors = neighbors.values
    weights = weights.values
    num_umi = num_umi.values

    D = compute_node_degree(neighbors, weights)
    Wtot2 = (weights**2).sum()

    counts = create_centered_counts(counts, model, num_umi)

    def initializer():
        global g_neighbors
        global g_weights
        global g_Wtot2
        global g_D
        global g_counts
        g_counts = counts
        g_neighbors = neighbors
        g_weights = weights
        g_Wtot2 = Wtot2
        g_D = D

    pairs = list(itertools.combinations(range(counts.shape[0]), 2))

    if jobs > 1:

        with multiprocessing.Pool(
                processes=jobs, initializer=initializer) as pool:

            results = list(
                tqdm(
                    pool.imap(_map_fun_parallel_pairs_centered, pairs),
                    total=len(pairs)
                )
            )
    else:
        def _map_fun(rowpair):
            return _compute_hs_pairs_inner_centered(
                rowpair, counts, neighbors, weights, Wtot2, D)
        results = list(
            tqdm(
                map(_map_fun, pairs),
                total=len(pairs)
            )
        )

    N = counts.shape[0]
    pairs = np.array(pairs)
    vals_lc = np.array([x[0] for x in results])
    vals_z = np.array([x[1] for x in results])
    lcs = expand_pairs(pairs, vals_lc, N)
    lc_zs = expand_pairs(pairs, vals_z, N)

    lc_maxs = compute_local_cov_pairs_max(D, counts)
    lcs = lcs / lc_maxs

    lcs = pd.DataFrame(lcs, index=genes, columns=genes)
    lc_zs = pd.DataFrame(lc_zs, index=genes, columns=genes)

    return lcs, lc_zs


def compute_hs_pairs_centered_cond(counts, neighbors, weights,
                                   num_umi, model, jobs=1):

    genes = counts.index

    counts = counts.values
    neighbors = neighbors.values
    weights = weights.values
    num_umi = num_umi.values

    D = compute_node_degree(neighbors, weights)

    counts = create_centered_counts(counts, model, num_umi)

    eg2s = np.array(
        [
            conditional_eg2(counts[i], neighbors, weights)
            for i in range(counts.shape[0])
        ]
    )

    def initializer():
        global g_neighbors
        global g_weights
        global g_counts
        global g_eg2s
        g_counts = counts
        g_neighbors = neighbors
        g_weights = weights
        g_eg2s = eg2s

    pairs = list(itertools.combinations(range(counts.shape[0]), 2))

    if jobs > 1:

        with multiprocessing.Pool(
                processes=jobs, initializer=initializer) as pool:

            results = list(
                tqdm(
                    pool.imap(_map_fun_parallel_pairs_centered_cond, pairs),
                    total=len(pairs)
                )
            )
    else:
        def _map_fun(rowpair):
            return _compute_hs_pairs_inner_centered_cond_sym(
                rowpair, counts, neighbors, weights, eg2s)
        results = list(
            tqdm(
                map(_map_fun, pairs),
                total=len(pairs)
            )
        )

    N = counts.shape[0]
    pairs = np.array(pairs)
    vals_lc = np.array([x[0] for x in results])
    vals_z = np.array([x[1] for x in results])
    lcs = expand_pairs(pairs, vals_lc, N)
    lc_zs = expand_pairs(pairs, vals_z, N)

    lc_maxs = compute_local_cov_pairs_max(D, counts)
    lcs = lcs / lc_maxs

    lcs = pd.DataFrame(lcs, index=genes, columns=genes)
    lc_zs = pd.DataFrame(lc_zs, index=genes, columns=genes)

    return lcs, lc_zs


def _map_fun_parallel_pairs(row_i):
    global g_neighbors
    global g_weights
    global g_num_umi
    global g_model
    global g_centered
    global g_Wtot2
    global g_D
    global g_counts
    return _compute_hs_pairs_inner(
        row_i, g_counts, g_neighbors, g_weights, g_num_umi,
        g_model, g_centered, g_Wtot2, g_D)


def _map_fun_parallel_pairs_centered(rowpair):
    global g_neighbors
    global g_weights
    global g_num_umi
    global g_model
    global g_Wtot2
    global g_D
    global g_counts
    return _compute_hs_pairs_inner_centered(
        rowpair, g_counts, g_neighbors, g_weights, g_Wtot2, g_D)


def _map_fun_parallel_pairs_centered_cond(rowpair):
    global g_neighbors
    global g_weights
    global g_counts
    global g_eg2s
    return _compute_hs_pairs_inner_centered_cond_sym(
        rowpair, g_counts, g_neighbors, g_weights, g_eg2s
    )


@njit
def expand_pairs(pairs, vals, N):

    out = np.zeros((N, N))

    for i in range(len(pairs)):

        x = pairs[i, 0]
        y = pairs[i, 1]
        v = vals[i]

        out[x, y] = v
        out[y, x] = v

    return out


def compute_local_cov_pairs_max(node_degrees, counts):
    """
    For a Genes x Cells count matrix, compute the maximal pair-wise correlation
    between any two genes
    """

    N_GENES = counts.shape[0]

    gene_maxs = np.zeros(N_GENES)
    for i in range(N_GENES):
        gene_maxs[i] = compute_local_cov_max(node_degrees, counts[i])

    result = gene_maxs.reshape((-1, 1)) + gene_maxs.reshape((1, -1))
    result = result / 2
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hotspot']

package_data = \
{'': ['*']}

install_requires = \
['anndata>=0.7',
 'matplotlib>=3.0.0',
 'numba>=0.43.1',
 'numpy>=1.16.4',
 'pandas>=0.24.0',
 'pynndescent>=0.5',
 'scikit-learn>=0.21.2',
 'scipy>=1.2.1',
 'seaborn>=0.9.0',
 'statsmodels>=0.9.0',
 'tqdm>=4.32.2']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>1.0'],
 'docs': ['sphinx-book-theme>=0.2.0', 'nbsphinx', 'ipython'],
 'test': ['pytest>=5.0']}

setup_kwargs = {
    'name': 'hotspotsc',
    'version': '1.0',
    'description': 'Identifying informative genes (and gene modules) in a single-cell dataset.',
    'long_description': '![Build Status](https://github.com/YosefLab/hotspot/workflows/hotspot/badge.svg)\n# Hotspot\n\nHotspot is a tool for identifying informative genes (and gene modules) in a single-cell dataset.\n\nImportantly \'informative\' is decided based on how well a gene\'s variation agrees with some cell metric - some similarity mapping between cells.\n\nGenes which are informative are those whose expression varies in similar way among cells which are nearby in the given metric.\n\nThe choice of metric allows you to evaluate different types of gene modules:\n\n- Spatial:  For spatial single-cell datasets, you can define cell similarity by proximity in physical, 2D/3D space.  When set up this way, Hotspot can be used to identify **spatially-varying genes**.\n- Lineage:  For single-cell datasets with a lineage tracing system, you can infer a cell developmental lineage and use that as the cell metric.  Then Hotspot can be used to identify **heritable patterns of gene expression.**\n- Transcriptional:  The result of a dimensionality reduction procedure can be used create the similarity metric. With this configuration, Hotspot identifies gene modules with local correlation.  This can be particularly useful for interpreting axes of variation in non-linear dimensionality reduction procedures where the mapping between components and genes is not straightforward to evaluate.\n\n**Examples**\n\n- [Spatial Data (Mouse Cerebellum)](https://hotspot.readthedocs.io/en/latest/Spatial_Tutorial.html)\n- [Lineage Data (Mouse Embryogenesis)](https://hotspot.readthedocs.io/en/latest/Lineage_Tutorial.html)\n- [Human CD4 T cells](https://hotspot.readthedocs.io/en/latest/CD4_Tutorial.html)\n\n**Other Links**\n\n- [Publication](https://linkinghub.elsevier.com/retrieve/pii/S2405471221001149)\n- [bioRxiv Preprint](https://www.biorxiv.org/content/10.1101/2020.02.06.937805v1)\n    - [Associated analysis repo](https://github.com/deto/Hotspot_analysis)\n- [Documentation](https://hotspot.readthedocs.io/en/latest/)\n\n# Installation\n\nHotspot is installed directly from PyPI using the following command:\n\n```\npip install hotspotsc\n```\n\nHotspot can be imported as\n\n```\nimport hotspot\n```\n\n# Usage\n\nThe following steps are used when analyzing data in Hotspot:\n\n1. Create the Hotspot object\n2. Compute the KNN graph\n3. Find informative genes (by gene autocorrelation)\n4. Evaluate pair-wise gene associations (gene local correlations)\n5. Group genes into modules\n6. Compute summary per-cell module scores\n\nHere we describe each step in order:\n\n## Create the Hotspot object\n\nTo start an analysis, first create the hotspot object\nWhen creating the object, you need to specify:\n- The gene counts matrix\n- Which background model of gene expression to use\n- The metric space to use to evaluate cell-similarity\n- The per-cell scaling factor\n    - By default, number of umi per cell is used\n\nFor example:\n```\nimport hotspot\nhs = hotspot.Hotspot(\n    adata,\n    layer_key="counts",\n    model=\'danb\',\n    latent_obsm_key="X_pca",\n    umi_counts_obs_key="total_counts"\n)\n```\n\nIn the example above:\n- `adata` is a [AnnData](https://anndata.readthedocs.io/en/latest/) object of cells by genes\n- `layer_key` is the layer of `adata` containing count information\n- `model` \'danb\' selects the umi-adjusted negative binomial model\n- `latent_obsm_key` is the `.obsm` key of `adata` containing Cells x Components matrix (the PCA-reduced space)\n- `umi_counts_obs_key` is the `.obs` key of `adata` with the UMI count for each cell\n\n\n### Alternative choices for \'model\'\n\nThe model is used to fit per-cell expectations for each gene assuming no correlations. This is used as the null model when evaluating autocorrelation and gene-gene local correlations.  The choices are:\n\n- **danb**: \'Depth-adjusted negative binomial\' (aka NBDisp model) from [M3Drop](https://bioconductor.org/packages/release/bioc/html/M3Drop.html)\n- **bernoulli**: Here only the detection probability for each gene is estimated.  Logistic regression on gene bins is used to evaluate this per-gene and per-cell as a function of the cells `umi_count` value.\n- **normal**: Here expression values are assumed to be normally-distributed and scaled by the values in `umi_count`.\n- **none**: With this option, the values are assumed to be pre-standardized\n\n\n### Choosing different metrics\n\nAbove we used `latent` as the input option.  This assumes that cells are in an N-dimensional space and similarity between cells is evaluated by computing euclidean distances in this space.  Either the results of a dimensionality reduction or modeling procedure can be input here, or when working with spatial data, the per-cell coordinates.\n\nAlternately, instead of `latent`, you can specify either `tree` or `distances`.\n\n`tree` is used for a developmental lineage.  In this form, `tree` should be an `ete3.TreeNode` object representing the root of a Tree with each cell as its leaves.  This could be constructed programmatically (see ete3\'s documentation for details) or if your lineage is stored in a Newick file format, you can load it into an ete3.TreeNode object by running `ete3.Tree(\'my_newick.txt\')`.  Note: leaf nodes in the tree must have names that match the column labels in the `counts` input (e.g., cell barcodes).\n\n`distances` is used to specify cell-cell distances directly.  The value entered should be a Cells x Cells pandas DataFrame.\n\n\n## Compute the KNN graph\n\nThe K-nearest-neighbors graph is computed by running:\n\n```\nhs.create_knn_graph(weighted_graph=False, n_neighbors=30)\n```\n\nInput options:\n- `weighted_graph`: bool, whether or not the graph has weights (scaled by cell-cell distances) or is binary\n- `n_neighbors`:  the number of neighbors per cell to use. Larger neighborhood sizes can result in more robust detection of correlations and autocorrelations at a cost of missing more fine-grained, smaller-scale patterns and increasing run-time\n- `neighborhood_factor`: float, used when `weighted_graph=True`.  Weights decay proportionally to `exp(-distance^2/distance_N^2)` where `distance_N` is the distance to the `n_neighbors/neighborhood_factor`th neighbor.\n\nGenerally, the above defaults should be fine in most cases.\n\n\n## Find informative genes (by gene autocorrelation)\n\nTo compute per-gene autocorrelations, run:\n\n```\nhs_results = hs.compute_autocorrelations()\n```\n\nAn optional argument, `jobs` can be specified to invoke parallel jobs for a speedup on multi-core machines.\n\nThe output is a pandas DataFrame (and is also saved in `hs.results`):\n\n| Gene            |     C |      Z |     Pval |      FDR |\n|:----------------|------:|-------:|---------:|---------:|\n| ENSG00000139644 | 0.069 | 10.527 | 3.21e-26 | 7.45e-25 |\n| ENSG00000179218 | 0.071 | 10.521 | 3.43e-26 | 7.93e-25 |\n| ENSG00000196139 | 0.081 | 10.517 | 3.59e-26 | 8.28e-25 |\n| ENSG00000119801 | 0.062 | 10.515 | 3.68e-26 | 8.48e-25 |\n| ENSG00000233355 | 0.058 | 10.503 | 4.15e-26 | 9.55e-25 |\n|      ...        |  ...  |   ...  |    ...   |    ...   |\n\n\nColumns are:\n- `C`: Scaled -1:1 autocorrelation coeficients\n- `Z`: Z-score for autocorrelation\n- `Pval`:  P-values computed from Z-scores\n- `FDR`:  Q-values using the Benjamini-Hochberg procedure\n\n\n## Evaluate pair-wise gene associations (gene local correlations)\n\nTo group genes into modules, we need to first evaluate their pair-wise local correlations\n\nBetter than regular correlations, these \'local\' correlations also take into accounts associations where one gene, X, is expression \'near\' another gene Y in the map.  This can better resolve correlations between sparsely detected genes.\n\n```\nhs_genes = hs_results.loc[hs_results.FDR < 0.05].index # Select genes\n\nlocal_correlations = hs.compute_local_correlations(hs_genes, jobs=4) # jobs for parallelization\n```\n\nHere we run only on a subset of genes as evaluating all pair-wise correlations is very expensive computationally.   The autocorrelation ordering gives us a natural method to select the informative genes for this purpose.\n\nThe output is a genes x genes pandas DataFrame of Z-scores for the local correlation values between genes.  The output is also stored in `hs.local_correlation_z`.\n\n## Group genes into modules\nNow that pair-wise local correlations are calculated, we can group genes into modules.\n\nTo do this, a convenience method is included `create_modules` which performs\nagglomerative clustering.\n\n```\nmodules = hs.create_modules(\n    min_gene_threshold=30, core_only=True, fdr_threshold=0.05\n)\n```\n\nA note on the arguments - agglomerative clustering proceeds by joining together genes with the highest pair-wise Z-score with the following caveats:\n\n- If the FDR-adjusted p-value of the correlation between two branches exceeds `fdr_threshold`,\nthen the branches are not merged.\n- If two branches are two be merged and they are both have at least `min_gene_threshold` genes,\nthen the branches are not merged.  Further genes that would join to the resulting merged module\nsmaller average correlations between genes, i.e. the least-dense module (if `core_only=False`)\n\nThis method was used to preserved substructure (nested modules) while still giving the analyst\nsome control. However, since there are a lot of ways to do hierarchical clustering, you can also\nmanually cluster using the gene-distances in `hs.local_correlation_z`\n\nThe output is a pandas Series that maps gene to module number.  Unassigned genes are indicated with a module number of `-1`.  The output is also stored in `hs.modules`\n\nCorrelation modules can be visualized by running `hs.plot_local_correlations()`:\n\n<img src="https://hotspot.readthedocs.io/en/latest/_images/CD4_Tutorial_29_0.png" alt="Local Correlation Plot" width="500">\n\n\n## Compute summary per-cell module scores\n\nFinally, summary per-cell scores can be computed for a module.  This is useful for visualizng the general pattern of expression for genes in a module.\n\n```\nmodule_scores = hs.calculate_module_scores()\n```\n\nThe output is a pandas DataFrame (cells x modules) and is also saved in `hs.module_scores`\n\n`module_scores`:\n\n|                    |      1 |      2 |      3 |      4 |      5 |\n|:-------------------|-------:|-------:|-------:|-------:|-------:|\n| AAACCCAAGGCCTAGA-1 |  0.785 | -2.475 | -1.407 | -0.681 | -1.882 |\n| AAACCCAGTCGTGCCA-1 | -5.76  |  5.241 |  6.931 |  1.928 |  4.317 |\n| AAACCCATCGTGCATA-1 | -2.619 |  3.572 |  0.143 |  1.832 |  1.585 |\n| AAACGAAGTAATGATG-1 | -8.778 |  4.012 |  6.927 |  1.181 |  3.882 |\n| AAACGCTCATGCACTA-1 |  2.297 | -2.517 | -1.421 | -1.102 | -1.547 |\n|        ...         |   ...  |   ...  |   ...  |   ...  |   ...  |\n\nThese can then be plotted onto other visual representations of the cells.  For example, for spatial modules (from data in [Rodriques et al, 2019](https://www.ncbi.nlm.nih.gov/pubmed/30923225)) this looks like:\n\n<img src="https://hotspot.readthedocs.io/en/latest/_images/Spatial_Tutorial_23_0.png" alt="Spatial Gene Modules" width="500">\n',
    'author': 'David DeTomaso',
    'author_email': 'David.DeTomaso@berkeley.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yoseflab/hostpost',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
